/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2020-2021 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

/* Compiler headers */
#include <stdbool.h>

/* Early FSP headers */

/* Hardware headers */

/* Late FSP headers */
#include <misc/macros.h>                   // for START_RFD_BLOCK, END_RFD_BLOCK, INLINE_RFD
#include <misc/nvmisc_drf.h>               // for DRF macros

/* Module-specific FSP headers */
#include <cpu/csr.h>
#include <cpu/fence.h>
#include <cpu/io.h>
#include <cpu/shutdown.h>

START_RFD_BLOCK(MISRA, DEVIATE, Rule_1_2, "Approval: Bug 200531996, DR: SWE-FSP-009-SWSADR.docx",
                MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx")

__attribute__ ((noreturn))
__attribute__((__optimize__("no-stack-protector")))
void riscvPanic(void)
{
    localWrite(NV_PRGNLCL_RISCV_ICD_CMD, 0);
    while(true) {}
}

__attribute__ ((noreturn))
__attribute__((__optimize__("no-stack-protector")))
void riscvShutdown(void)
{
// Disable interrupts
    csr_write(NV_RISCV_CSR_MIE, 0);
    riscvLwfenceIO();
// Halt
#if defined(NV_RISCV_CSR_MOPT)
    __asm__ volatile ("csrrwi zero, %0, %1" : :"i"(NV_RISCV_CSR_MOPT), "i"( DRF_DEF64(_RISCV, _CSR_MOPT, _CMD, _HALT)));
#endif
    while(true) {}
}

END_RFD_BLOCK(MISRA, DEVIATE, Rule_1_2, "Approval: Bug 200531996, DR: SWE-FSP-009-SWSADR.docx",
              MISRA, DEVIATE, Directive_4_9, "Approval: Bug 200531995, DR: SWE-FSP-012-SWSADR.docx")
